from flask import jsonify, request, Flask, make_response
import logging
import re
from multiprocessing import Process
import requests
import json
import socket
import time
from collections import OrderedDict
from tlib.base.NetworkHelper import NetworkHelper
import base64

mock_server = Flask(__name__)
_rules = OrderedDict()
_req_urls = []
_requests = []


def generate_error_response(message, status_code):
    """
    Method to build an error response and log it
    """
    logging.error(message)
    logging.error(request.data)

    r = jsonify(status=message)
    r.status_code = status_code
    return r


@mock_server.route("/mock/shutdown", methods=['GET'])
def shutdown():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()
    return jsonify(status='Server is shutting down')


@mock_server.route("/mock/responses", methods=['POST'])
def add_response():
    """
    This method adds new responses to the mock.
    To add a response send a POST request with a payload like this:

    {
        "url_filter": ".*",
        "headers": {
            "Accept": "text/xml"
        },
        "body": "Sample body",
        "status_code": 200
    }

    Server will validate each matching rule and apply the first match
    If there is no match, it will return a 500 response

    To add binary responses, send data encoded in base64 and add this value to the request
    "encoding": "base64".
    For example:

    {
        "url_filter": "/file.zip",
        "headers": {
            "Accept": "application/zip"
        },
        "body": "ADRFDRRFJFJJSMMSK",
        "encoding": "base64",
        "status_code": 200
    }

    """
    try:
        payload = request.get_json(force=True)
    except Exception as e:
        return generate_error_response("Payload is not a valid JSON string.\n%s" % e.message, 400)

    if not isinstance(payload, dict):
        return generate_error_response("Payload is not a valid JSON string", 400)

    # Parse data from request
    if "url_filter" in payload.keys():
        key = payload["url_filter"]
        # Check filter is not empty
        if key == "":
            return generate_error_response("url_filter can't be an empty string", 400)

        try:
            url_filter = re.compile(payload["url_filter"])
        except Exception as e:
            return generate_error_response("url_filter is not a valid regular expression:\\n%s" % e.message, 400)
    else:
        return generate_error_response("url_filter is a required parameter", 400)

    if "headers" in payload.keys():
        if type(payload["headers"]) is dict:
            headers = payload["headers"]
        else:
            return generate_error_response("headers is not a dictionary:\\n%s" % payload["headers"], 400)
    else:
        headers = {}

    if "body" in payload.keys():
        if "encoding" in payload.keys():
            # Payload can be plain text or encoded in base64
            if payload["encoding"] not in ("base64", "text"):
                return generate_error_response("Encoding '%s' is not valid. Supported encodings are text and base64" %
                                               payload["encoding"], 400)

            if payload["encoding"] == "base64":
                body = base64.b64decode(payload["body"])
            else:
                body = payload["body"]
        else:
            # If encoding is not specified, assume it's plain text
            body = payload["body"]
    else:
        body = ""

    if "status_code" in payload.keys():
        if not isinstance(payload["status_code"], int):
            return generate_error_response("Status code is not an integer: %s" % payload["status_code"], 400)

        status_code = payload["status_code"]
    else:
        status_code = 200

    # Save parsed data
    new_rule = {"url_filter": url_filter, "headers": headers, "body": body, "status_code": status_code}
    _rules[key] = new_rule

    return jsonify(status="OK")


@mock_server.route("/mock/responses", methods=['DELETE'])
def clear_responses():
    """
    Delete existing responses
    """
    _rules.clear()
    return jsonify(status="All rules were deleted")


@mock_server.route("/mock/responses", methods=['GET'])
def get_responses():
    """
    Get all responses
    """
    rules_as_text = []
    for rule in _rules.values():
        # make a copy so we don't modify original
        rule = rule.copy()

        # Convert regex to str
        rule["url_filter"] = rule["url_filter"].pattern

        # Add rule to list
        rules_as_text.append(rule)

    return jsonify(rules=rules_as_text)


@mock_server.route("/mock/requests", methods=['DELETE'])
def clear_requests():
    """
    Delete existing requests
    """
    while _req_urls:
        _req_urls.pop()

    while _requests:
        _requests.pop()

    return jsonify(status="All requests were deleted")


@mock_server.route("/mock/requests", methods=['GET'])
def get_requests():
    """
    Get all requests.
    If parameter version is equal to 2, a detailed request (including headers and body) will be returned, otherwise
    just the URL will be returned
    """
    if request.args.get('version') == '2':
        return jsonify(requests=_requests)
    else:
        return jsonify(requests=_req_urls)


@mock_server.route('/', defaults={'path': ''}, methods=['GET', 'POST', 'PUT', 'DELETE'])
@mock_server.route('/<path:path>', methods=['GET', 'POST', 'PUT', 'DELETE'])
def catch_all(path):
    """
    This method will catch all requests for which there are no explicit routes.
    Here is where we build responses based on the rules that have been configured
    It will go though the list of rules and apply one by one until a match is found.
    If there is no match, it will return a 500 response
    """
    _req_urls.append(request.url)

    headers = {}
    for h in str(request.headers).split("\r\n"):
        if h:
            parts = h.split(": ")
            h_name = parts[0]
            h_value = ":".join(parts[1:])
            headers[h_name] = h_value

    # Store full request
    req = {
        'Request': {
            "Scheme": request.scheme,
            "Host": NetworkHelper.get_hostname_from_url(request.url),
            "Port": NetworkHelper.get_port_from_url(request.url),
            "Path": request.path,
            "Args": dict(request.args),
            "Method": request.method,
            "Headers": headers,
            "Body": request.data
        }
    }

    _requests.append(req)

    for rule in _rules.values():
        regex = rule["url_filter"]
        if regex.search("/" + path):
            r = make_response()
            r .headers = rule["headers"]
            r .data = rule["body"]
            r .status_code = rule["status_code"]
            return r

    # Default values returned when there wasn't a match in the rules
    return generate_error_response("Mock has not been configured", 500)


class ApiMockServer(Process):
    """
    Helper Class to interact with the ApiMockServer
    Provides methods to add, remove mock objects as well as incoming requests
    """

    server = mock_server

    def __init__(self, port):
        super(ApiMockServer, self).__init__()

        self.port = port
        self.host = "0.0.0.0"  # host IP to be server by the Flask server
        self.localhost = self.host_ip()  # IP to communicate with server
        self.base_url = "http://{host}:{port}".format(host=self.localhost, port=self.port)
        self.responses_url = self.base_url + "/mock/responses"
        self.requests_url = self.base_url + "/mock/requests"
        self.shutdown_url = self.base_url + "/mock/shutdown"

    def run(self):
        """
        start the mock server
        """
        super(ApiMockServer, self).run()
        self.server.run(host=self.host, port=self.port)

    def start(self):
        super(ApiMockServer, self).start()
        self._wait_for_server_to_start()

    def stop(self):
        """
        Shutdown the server and terminate process
        :return:
        """
        self._wait_for_server_to_start()
        requests.get(self.shutdown_url)
        self.terminate()

    def add_response(self, url_filter=None, status_code=200, headers=None, body="", encoding="text", data=None):
        """
        Add the response to the mock server

        :param url_filter: Regular expression to match the url path
        :type url_filter: str
        :param status_code: Expected status code
        :type status_code: int
        :param headers: Headers to return in the response
        :type headers: dict
        :param body: response body as plain text
        :type body: str
        :param encoding: If body is binary, this should be set to string "base64".
        Otherwise, leave empty or set to "text"
        :type encoding: str
        :param data: If provided, it will be the rule to add to the mock.
        :type data: str
        :return: Mock server response
        """
        self._wait_for_server_to_start()

        if not data:
            data = {"url_filter": url_filter,
                    "status_code": status_code,
                    "headers": headers if headers else {},
                    "body": body,
                    "encoding": encoding}

        return requests.post(self.responses_url, data=json.dumps(data))

    def add_binary_response(self, url_filter=None, status_code=200, headers=None, body=""):
        """
        Adds a binary response to the mock server

        :param url_filter: Regular expression to match the url path
        :type url_filter: str
        :param status_code: Expected status code
        :type status_code: int
        :param headers: Headers to return in the response
        :type headers: dict
        :param body: response body as plain text
        :type body: str
        :return: Mock server response
        """
        return self.add_response(url_filter, status_code=status_code, headers=headers, encoding="base64",
                                 body=base64.b64encode(body))

    def add_text_response(self, url_filter=None, status_code=200, headers=None, body=""):
        """
        Adds a binary response to the mock server

        :param url_filter: Regular expression to match the url path
        :type url_filter: str
        :param status_code: Expected status code
        :type status_code: int
        :param headers: Headers to return in the response
        :type headers: dict
        :param body: response body as plain text
        :type body: str
        :return: Mock server response
        """
        return self.add_response(url_filter, status_code=status_code, headers=headers, body=body)

    def get_responses(self):
        """
        Returns all the responses stored on the MockServer
        :return:
        """
        self._wait_for_server_to_start()
        response = requests.get(self.responses_url)
        return response if response else None

    def clear_responses(self):
        """
        Delete all the responses stored on the MockServer
        :return:
        """
        self._wait_for_server_to_start()
        return requests.delete(self.responses_url)

    def get_requests(self, version=1):
        """
        Returns all the requests stored on the MockServer
        :param version:  If parameter is equal to 2, a detailed request (including headers and body) will be returned,
        otherwise just the URL will be returned
        :return:
        """
        self._wait_for_server_to_start()
        response = requests.get(self.requests_url + "?version=" + version)
        if response:
            return response
        return None

    def clear_requests(self):
        """
        Delete all the requests stored on the MockServer
        :return:
        """
        self._wait_for_server_to_start()
        return requests.delete(self.requests_url)

    def host_ip(self):
        """
        Dirty hack to return the localhost IP,
        works only when you have an internet connection
        :return:
        """
        return [(s.connect(('8.8.8.8', 80)), s.getsockname()[0], s.close())
                for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]

    def _is_server_running(self):
        """
        check if the server is running and port is open
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((self.host_ip(), self.port))
        return result == 0

    def _wait_for_server_to_start(self):
        """
        Latency to avoid the access when the server is not started yet
        """
        while not self._is_server_running():
            time.sleep(0.1)

    def get_base_url(self):
        return self.base_url

if __name__ == '__main__':
    # to use as standalone application
    # run 'python ApiMockServer.py'
    server = ApiMockServer(port=10000)
    server.start()
